"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const logger_1 = require("./logger");
const PropTree_1 = require("./PropTree");
const rgxObjectTokenize = /(<|>|,)/;
const rgxCommaAll = /,/g;
const rgxParensAll = /\(|\)/g;
const anyTypeNode = ts.createKeywordTypeNode(ts.SyntaxKind.AnyKeyword);
const strTypeNode = ts.createKeywordTypeNode(ts.SyntaxKind.StringKeyword);
const anyGeneric = { kind: 'type', name: 'any', resolved: anyTypeNode };
const strGeneric = { kind: 'type', name: 'string', resolved: strTypeNode };
function toKeywordTypeKind(k) {
    if (!k || k.length === 0)
        return null;
    k = k.toUpperCase();
    switch (k) {
        case 'ANY': return ts.SyntaxKind.AnyKeyword;
        case 'UNKNOWN': return ts.SyntaxKind.UnknownKeyword;
        case 'NUMBER': return ts.SyntaxKind.NumberKeyword;
        case 'BIGINT': return ts.SyntaxKind.BigIntKeyword;
        case 'OBJECT': return ts.SyntaxKind.ObjectKeyword;
        case 'BOOLEAN': return ts.SyntaxKind.BooleanKeyword;
        case 'BOOL': return ts.SyntaxKind.BooleanKeyword;
        case 'STRING': return ts.SyntaxKind.StringKeyword;
        case 'SYMBOL': return ts.SyntaxKind.SymbolKeyword;
        case 'THIS': return ts.SyntaxKind.ThisKeyword;
        case 'VOID': return ts.SyntaxKind.VoidKeyword;
        case 'UNDEFINED': return ts.SyntaxKind.UndefinedKeyword;
        case 'NULL': return ts.SyntaxKind.NullKeyword;
        case 'NEVER': return ts.SyntaxKind.NeverKeyword;
        default:
            return null;
    }
}
exports.toKeywordTypeKind = toKeywordTypeKind;
function resolveOptional(doclet) {
    if (doclet.defaultvalue || doclet.optional)
        return ts.createToken(ts.SyntaxKind.QuestionToken);
    return undefined;
}
exports.resolveOptional = resolveOptional;
function resolveVariable(doclet) {
    if (doclet.variable)
        return ts.createToken(ts.SyntaxKind.DotDotDotToken);
    return undefined;
}
exports.resolveVariable = resolveVariable;
function getExprWithTypeArgs(identifier) {
    const expr = ts.createIdentifier(identifier);
    return ts.createExpressionWithTypeArguments(undefined, expr);
}
function resolveHeritageClauses(doclet, onlyExtend) {
    const clauses = [];
    let extensions = doclet.augments || [];
    if (onlyExtend) {
        extensions = extensions.concat(doclet.implements || []);
        extensions = extensions.concat(doclet.mixes || []);
    }
    if (extensions.length) {
        clauses.push(ts.createHeritageClause(ts.SyntaxKind.ExtendsKeyword, extensions.map(getExprWithTypeArgs)));
    }
    if (onlyExtend)
        return clauses;
    let implementations = (doclet.implements || []).concat(doclet.mixes || []);
    if (implementations.length) {
        clauses.push(ts.createHeritageClause(ts.SyntaxKind.ImplementsKeyword, implementations.map(getExprWithTypeArgs)));
    }
    return clauses;
}
exports.resolveHeritageClauses = resolveHeritageClauses;
function resolveTypeParameters(doclet) {
    const typeParams = [];
    if (doclet.tags) {
        for (let i = 0; i < doclet.tags.length; ++i) {
            const tag = doclet.tags[i];
            if (tag.title === 'template') {
                const types = (tag.text || 'T').split(',');
                for (let x = 0; x < types.length; ++x) {
                    const name = types[x].trim();
                    if (!name)
                        continue;
                    typeParams.push(ts.createTypeParameterDeclaration(name, undefined, undefined));
                }
            }
        }
    }
    return typeParams;
}
exports.resolveTypeParameters = resolveTypeParameters;
function resolveType(t, doclet) {
    if (!t || !t.names || t.names.length === 0) {
        if (doclet && doclet.properties)
            return resolveTypeName('object', doclet);
        if (doclet) {
            logger_1.warn(`Unable to resolve type for ${doclet.longname || doclet.name}, none specified in JSDoc. Defaulting to \`any\`.`, doclet);
        }
        else {
            logger_1.warn(`Unable to resolve type for an unnamed item, this is likely due to invalid JSDoc.` +
                ` Often this is caused by invalid JSDoc on a parameter. Defaulting to \`any\`.`, doclet);
        }
        return anyTypeNode;
    }
    if (t.names.length === 1) {
        return resolveTypeName(t.names[0], doclet);
    }
    else {
        const types = [];
        for (let i = 0; i < t.names.length; ++i) {
            types.push(resolveTypeName(t.names[i], doclet));
        }
        return ts.createUnionTypeNode(types);
    }
}
exports.resolveType = resolveType;
function resolveTypeName(name, doclet) {
    if (!name) {
        logger_1.warn('Unable to resolve type name, it is null, undefined, or empty. Defaulting to \`any\`.', doclet);
        return anyTypeNode;
    }
    if (name === '*')
        return anyTypeNode;
    const keyword = toKeywordTypeKind(name);
    if (keyword !== null) {
        if (keyword === ts.SyntaxKind.ThisKeyword)
            return ts.createThisTypeNode();
        if (keyword === ts.SyntaxKind.ObjectKeyword) {
            if (!doclet || !doclet.properties)
                return anyTypeNode;
            else
                return resolveTypeLiteral(doclet.properties);
        }
        return ts.createKeywordTypeNode(keyword);
    }
    const upperName = name.toUpperCase();
    if (upperName === 'FUNCTION' || upperName === 'FUNCTION()') {
        if (doclet && doclet.kind === 'typedef') {
            const params = createFunctionParams(doclet);
            const type = createFunctionReturnType(doclet);
            return ts.createFunctionTypeNode(undefined, params, type);
        }
        else {
            const anyArray = ts.createArrayTypeNode(anyTypeNode);
            const dotDotDot = ts.createToken(ts.SyntaxKind.DotDotDotToken);
            const param = ts.createParameter(undefined, undefined, dotDotDot, 'params', undefined, anyArray, undefined);
            return ts.createFunctionTypeNode(undefined, [param], anyTypeNode);
        }
    }
    if (upperName.indexOf('.<') !== -1) {
        return resolveGenericType(name);
    }
    if (name.indexOf('|') !== -1) {
        const nameParts = name.split('|');
        const types = [];
        for (let i = 0; i < nameParts.length; ++i) {
            const subName = nameParts[i].replace(rgxParensAll, '').trim();
            types.push(resolveTypeName(subName, doclet));
        }
        return ts.createUnionTypeNode(types);
    }
    return ts.createTypeReferenceNode(name, undefined);
}
exports.resolveTypeName = resolveTypeName;
function resolveGenericType(name) {
    const parts = name.split(rgxObjectTokenize);
    if (parts.length <= 1) {
        logger_1.warn('Invalid object type encountered, this is likely due to invalid JSDoc. Defaulting to \`any\`.', name);
        return anyTypeNode;
    }
    let lastObj = null;
    let parentStack = [];
    for (let i = 0; i < parts.length; ++i) {
        const partName = parts[i].trim();
        if (!partName || partName === ',')
            continue;
        if (partName.endsWith('.')) {
            const parent = parentStack[parentStack.length - 1];
            lastObj = {
                kind: 'generic',
                name: partName.substr(0, partName.length - 1),
                parent,
                types: [],
            };
            if (parent)
                parent.types.push(lastObj);
        }
        else if (partName === '<') {
            if (lastObj && lastObj.kind === 'generic')
                parentStack.push(lastObj);
        }
        else if (partName === '>') {
            parentStack.pop();
        }
        else {
            const parent = parentStack[parentStack.length - 1];
            if (!parent) {
                logger_1.warn(`Invalid generic format encountered when parsing type: ${name}`);
                continue;
            }
            lastObj = {
                kind: 'type',
                name: partName,
                parent,
                resolved: resolveTypeName(partName),
            };
            parent.types.push(lastObj);
        }
    }
    if (!lastObj || !lastObj.parent) {
        logger_1.warn('Unable to resolve Object/Array complex type. Defaulting to \`any\`.', name);
        return anyTypeNode;
    }
    return resolveGenericTypeTree(lastObj.parent);
}
exports.resolveGenericType = resolveGenericType;
function resolveGenericTypeTree(bottom) {
    let lastType = null;
    let parent = bottom;
    while (parent) {
        if (parent.kind === 'generic') {
            const upperName = parent.name.toUpperCase();
            if (upperName === 'OBJECT') {
                let keyType = parent.types[0];
                if (!keyType) {
                    logger_1.warn(`Unable to resolve object key type, this is likely due to invalid JSDoc. Defaulting to \`string\`.`);
                    keyType = strGeneric;
                }
                else if (keyType.kind !== 'type'
                    || (keyType.name !== 'string' && keyType.name !== 'number')) {
                    logger_1.warn(`Invalid object key type. It must be \`string\` or \`number\`, but got: ${keyType.name}. Defaulting to \`string\`.`);
                    keyType = strGeneric;
                }
                let valType = parent.types[1];
                if (!valType) {
                    logger_1.warn('Unable to resolve object value type, this is likely due to invalid JSDoc. Defaulting to \`any\`.', parent);
                    valType = anyGeneric;
                }
                const indexParam = ts.createParameter(undefined, undefined, undefined, 'key', undefined, keyType.resolved, undefined);
                if (!valType.resolved) {
                    logger_1.warn('Unable to resolve object value type, this is likely a bug. Defaulting to \`any\`.', parent);
                    valType.resolved = anyTypeNode;
                }
                const indexSignature = ts.createIndexSignature(undefined, undefined, [indexParam], valType.resolved);
                lastType = parent.resolved = ts.createTypeLiteralNode([indexSignature]);
            }
            else if (upperName === 'ARRAY') {
                let valType = parent.types[0];
                if (!valType) {
                    logger_1.warn('Unable to resolve array value type, defaulting to \`any\`.', parent);
                    valType = anyGeneric;
                }
                if (!valType.resolved) {
                    logger_1.warn('Unable to resolve array value type, defaulting to \`any\`.', parent);
                    valType.resolved = anyTypeNode;
                }
                lastType = parent.resolved = ts.createArrayTypeNode(valType.resolved);
            }
            else if (upperName === 'CLASS') {
                let valType = parent.types[0];
                if (!valType) {
                    logger_1.warn('Unable to resolve array value type, defaulting to \`any\`.', parent);
                    valType = anyGeneric;
                }
                if (!valType.resolved) {
                    logger_1.warn('Unable to resolve array value type, defaulting to \`any\`.', parent);
                    valType.resolved = anyTypeNode;
                }
                lastType = parent.resolved = ts.createTypeQueryNode(ts.createIdentifier(valType.name));
            }
            else {
                const typeNodes = [];
                for (let i = 0; i < parent.types.length; ++i) {
                    let valType = parent.types[i];
                    if (!valType) {
                        logger_1.warn('Unable to resolve generic type parameter, defaulting to \`any\`.', parent);
                        valType = anyGeneric;
                    }
                    if (!valType.resolved) {
                        logger_1.warn('Unable to resolve generic type parameter, defaulting to \`any\`.', parent);
                        valType.resolved = anyTypeNode;
                    }
                    typeNodes.push(valType.resolved);
                }
                lastType = parent.resolved = ts.createTypeReferenceNode(parent.name, typeNodes);
            }
        }
        parent = parent.parent;
    }
    if (lastType === null)
        return anyTypeNode;
    return lastType;
}
function resolveTypeLiteral(props) {
    if (!props)
        return ts.createTypeLiteralNode([]);
    const tree = new PropTree_1.PropTree(props);
    return createTypeLiteral(tree.roots);
}
exports.resolveTypeLiteral = resolveTypeLiteral;
function createTypeLiteral(nodes) {
    const members = [];
    for (let i = 0; i < nodes.length; ++i) {
        const node = nodes[i];
        const opt = node.prop.optional ? ts.createToken(ts.SyntaxKind.QuestionToken) : undefined;
        const t = node.children.length ? createTypeLiteral(node.children) : resolveType(node.prop.type);
        members.push(ts.createPropertySignature(undefined, node.name, opt, t, undefined));
    }
    return ts.createTypeLiteralNode(members);
}
exports.createTypeLiteral = createTypeLiteral;
function createFunctionParams(doclet) {
    if (!doclet.params || !doclet.params.length)
        return [];
    const params = [];
    const tree = new PropTree_1.PropTree(doclet.params);
    if ((doclet.kind === 'function' || doclet.kind === 'typedef') && doclet.this) {
        const type = resolveType({ names: [doclet.this] }, doclet);
        params.push(ts.createParameter(undefined, undefined, undefined, 'this', undefined, type, undefined));
    }
    for (let i = 0; i < tree.roots.length; ++i) {
        const node = tree.roots[i];
        const opt = resolveOptional(node.prop);
        const dots = resolveVariable(node.prop);
        let type = node.children.length ? createTypeLiteral(node.children) : resolveType(node.prop.type);
        if (dots) {
            type = ts.createArrayTypeNode(type);
        }
        params.push(ts.createParameter(undefined, undefined, dots, node.name, opt, type, undefined));
    }
    return params;
}
exports.createFunctionParams = createFunctionParams;
function createFunctionReturnType(doclet) {
    if (doclet.returns && doclet.returns.length === 1) {
        return resolveType(doclet.returns[0].type, doclet);
    }
    else {
        return ts.createKeywordTypeNode(ts.SyntaxKind.VoidKeyword);
    }
}
exports.createFunctionReturnType = createFunctionReturnType;
//# sourceMappingURL=type_resolve_helpers.js.map